#ifndef _NF_SET_H
#define _NF_SET_H

/* Copyright (C) 2000-2002 Joakim Axelsson <gozem@linux.nu>
 *                         Patrick Schaaf <bof@bof.de>
 *                         Martin Josefsson <gandalf@wlug.westbo.se>
 * Copyright (C) 2003-2007 Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  
 */

#if 1
#define NF_SET_DEBUG
#endif

/* 
 * Used so that the kernel module and nfset-binary can match their versions 
 */
#define NF_SET_PROTOCOL_VERSION 2

#define NF_SET_MAXNAMELEN 32	/* set names and set typenames */
#define NFSET_MAXNAMELEN  NF_SET_MAXNAMELEN

/* Sets are identified by an id in kernel space. Tweak with nf_set_id_t
 * and NF_SET_INVALID_ID if you want to increase the max number of sets.
 */
typedef u_int16_t nf_set_id_t;

#define NF_SET_INVALID_ID	65535

/*
 * Option flags for kernel operations (ipt_set_info)
 */
#define NFSET_SRC 		0x01	/* Source match/add */
#define NFSET_DST		0x02	/* Destination match/add */
#define NFSET_MATCH_INV		0x04	/* Inverse matching */

/*
 * Set features
 */
#define NFSET_TYPE_IP		0x01	/* IP address type of set */
#define NFSET_TYPE_PORT		0x02	/* Port type of set */
#define NFSET_DATA_FIRST	0x04	/* Single data storage */
#define NFSET_DATA_SECOND	0x08	/* Double data storage */
#define NFSET_DATA_THIRD	0x10	/* Triple data storage */

#ifdef __KERNEL__

#define nf_set_printk(format, args...) 			\
	do {							\
		printk("%s: %s: ", __FILE__, __FUNCTION__);	\
		printk(format "\n" , ## args);			\
	} while (0)

#if defined(NF_SET_DEBUG)
#define DSET(format, args...) 					\
	do {							\
		printk("%s: %s (DBG): ", __FILE__, __FUNCTION__);\
		printk(format "\n" , ## args);			\
	} while (0)
#define NF_SET_ASSERT(x)					\
	do {							\
		if (!(x))					\
			printk("NF_SET_ASSERT: %s:%i(%s)\n",	\
				__FILE__, __LINE__, __FUNCTION__); \
	} while (0)
#else
#define DSET(format, args...)
#define NF_SET_ASSERT(x)
#endif

struct nf_set;

/*
 * The nf_set_type definition - one per set type, e.g. "ipmap".
 *
 * Each individual set has a pointer, set->type, going to one
 * of these structures. Function pointers inside the structure implement
 * the real behaviour of the sets.
 *
 * If not mentioned differently, the implementation behind the function
 * pointers of a set_type, is expected to return 0 if ok, and a negative
 * errno (e.g. -EINVAL) on error.
 */
struct nf_set_type {
	struct list_head list;	/* next in list of set types */

	/* test for IP in set (kernel: iptables -m set src|dst)
	 * return 0 if not in set, 1 if in set.
	 */
	int (*testip_kernel) (struct nf_set *set,
			      const struct sk_buff * skb,
			      const u_int32_t flags);

	/* Add IP into set (kernel: iptables ... -j SET set src|dst)
	 * Return -EEXIST if the address is already in the set,
	 * and -ERANGE if the address lies outside the set bounds.
	 * If the address was not already in the set, 0 is returned.
	 */
	int (*addip_kernel) (struct nf_set *set,
			     const struct sk_buff * skb,
			     const u_int32_t flags);

	/* remove IP from set (kernel: iptables ... -j SET set src|dst)
	 * Return -EEXIST if the address is NOT in the set,
	 * and -ERANGE if the address lies outside the set bounds.
	 * If the address really was in the set, 0 is returned.
	 */
	int (*delip_kernel) (struct nf_set *set,
			     const struct sk_buff * skb,
			     const u_int32_t flags);

	/* test for IP in set (userspace: nfset -T set IP)
	 * return 0 if not in set, 1 if in set.
	 */
	int (*testip) (struct nf_set *set,
		       struct nfattr *attr);

	/* Add IP into set (userspace: nfset -A set IP)
	 * Return -EEXIST if the address is already in the set,
	 * and -ERANGE if the address lies outside the set bounds.
	 * If the address was not already in the set, 0 is returned.
	 */
	int (*addip) (struct nf_set *set, 
		      struct nfattr *attr);

	/* remove IP from set (userspace: nfset -D set --entry x)
	 * Return -EEXIST if the address is NOT in the set,
	 * and -ERANGE if the address lies outside the set bounds.
	 * If the address really was in the set, 0 is returned.
	 */
	int (*delip) (struct nf_set *set, 
		      struct nfattr *attr);

	/* new set creation - allocated type specific items
	 */
	int (*create) (struct nf_set *set,
		       struct nfattr *attr);

	/* retry the operation after successfully tweaking the set
	 */
	int (*retry) (struct nf_set *set);

	/* set destruction - free type specific items
	 * There is no return value.
	 * Can be called only when child sets are destroyed.
	 */
	int (*destroy) (struct nf_set *set);

	/* set flushing - reset all bits in the set, or something similar.
	 * There is no return value.
	 */
	void (*flush) (struct nf_set *set);

	/* Listing: Get the header of the set, i.e. by which
	 * parameters was it created.
	 */
	int (*list_header) (struct sk_buff *skb,
			    const struct nf_set *set);

	/* Listing: Get the set members
	 *
	 * Fill in the information in the netlink message
	 */
	int (*list_members) (struct sk_buff *skb,
			     struct netlink_callback *cb,
			     const struct nf_set *set);

	/* Listing: netlink callback 'done' function
	 */
	int (*list_members_done) (struct netlink_callback *cb);

	/* Check whether the set members are busy: only for union
	 */
	int (*referenced) (struct nf_set *set);

	char typename[NF_SET_MAXNAMELEN];
	unsigned char features;
	int protocol_version;
	int family;

	/* Set this to THIS_MODULE if you are a module, otherwise NULL */
	struct module *me;
};

extern int nf_set_register_set_type(struct nf_set_type *set_type);
extern void nf_set_unregister_set_type(struct nf_set_type *set_type);

/* A generic nfset */
struct nf_set {
	char name[NF_SET_MAXNAMELEN];	/* the name of the set */
	rwlock_t lock;			/* lock for concurrency control */
	atomic_t ref;			/* in kernel references */
	struct nf_set_type *type; 	/* the set types */
	void *data;			/* pooltype specific data */
};

/* register and unregister set references */
extern nf_set_id_t nf_set_get_byname(const char name[NF_SET_MAXNAMELEN]);
extern nf_set_id_t nf_set_get_byindex(nf_set_id_t id);
extern void nf_set_put(nf_set_id_t id);

extern void nfset_put(nf_set_id_t index);

/* API for iptables set match, and SET target */
extern void nf_set_addip_kernel(nf_set_id_t id,
				const struct sk_buff *skb,
				const u_int32_t flags);
extern void nf_set_delip_kernel(nf_set_id_t id,
				const struct sk_buff *skb,
				const u_int32_t flags);
extern int nf_set_testip_kernel(nf_set_id_t id,
				const struct sk_buff *skb,
				const u_int32_t flags);

#endif				/* __KERNEL__ */

#endif /*_NF_SET_H*/
