#ifndef __NF_SET_IPMAP_H
#define __NF_SET_IPMAP_H

#ifdef NFSET_IPV6
#include <linux/in6.h>
#include <net/ipv6.h>
#define nf_set_ip_t		struct in6_addr
#define NF_SET_IP_FMT		NIP6_FMT
#define NF_SET_IP_PRT		NIP6
#define nf_set_type_ipmap	nf_set_type_ipmap_ipv6
#define AF_FAMILY		AF_INET6
#else /* IPv4 */
#include <linux/ip.h>
#define nf_set_ip_t		u_int32_t
#define NF_SET_IP_FMT		NIPQUAD_FMT
#define NF_SET_IP_PRT		HIPQUAD
#define nf_set_type_ipmap	nf_set_type_ipmap_ipv4
#define AF_FAMILY		AF_INET
#endif

/* Common definitions */
struct nf_set_ipmap {
	void *members;			/* the ipmap proper */
	unsigned int timeout;		/* timeout, if used */
	nf_set_ip_t first_ip;		/* host byte order, included in range */
	nf_set_ip_t last_ip;		/* host byte order, included in range */
	nf_set_ip_t netmask;		/* subnet netmask */
	u_int32_t ipnum;		/* number of IPs in the set */
	u_int8_t subnet;		/* subnet netmask */
};

/* Type speficic create data */
enum nfset_attr_create_ipmap_type {
	NFSET_CREATE_IPMAP_UNSPEC,
	NFSET_CREATE_IPMAP_IP,		/* IP address */
	NFSET_CREATE_IPMAP_NETMASK,	/* Netmask */
	NFSET_CREATE_IPMAP_SUBNET,	/* Subnet blocks in network */
	NFSET_CREATE_IPMAP_TIMEOUT,	/* Timeout */
	__NFSET_CREATE_IPMAP_MAX
};
#define NFSET_CREATE_IPMAP_MAX (__NFSET_CREATE_IPMAP_MAX - 1)

/* Type speficic adt data */
enum nfset_attr_adt_ipmap_type {
	NFSET_ADT_IPMAP_UNSPEC,
	NFSET_ADT_IPMAP_IP,		/* IP address */
	NFSET_ADT_IPMAP_TIMEOUT,	/* Timeout */
	__NFSET_ADT_IPMAP_MAX
};
#define NFSET_ADT_IPMAP_MAX (__NFSET_ADT_IPMAP_MAX - 1)

#define DSET_ADT(ip, orig) 					\
	DSET("ip:" NF_SET_IP_FMT " (" NF_SET_IP_FMT ")",	\
	     NF_SET_IP_PRT(*(ip)), NF_SET_IP_PRT(*(orig)))

static inline int bitmap_bytes(u_int32_t a, u_int32_t b)
{
	return 4 * ((((b - a + 8) / 8) + 3) / 4);
}

#ifdef NFSET_IPV6

static inline u_int32_t
ip_to_id(const struct nf_set_ipmap *map, struct in6_addr *ip)
{
	unsigned int i;
	
	for (i = 0; i < 4; i++)
		if (ip->s6_addr32[i] != map->first_ip.s6_addr32[i])
			break;
		
	return (ip->s6_addr32[i] - map->first_ip.s6_addr32[i]) >> (128 - map->subnet);
}

static inline int
ip_out_of_range(const struct nf_set_ipmap *map, struct in6_addr *ip)
{
	int i;
	
	for (i = 0; i < 4; i++)
		if ((ip->s6_addr32[i] != map->first_ip.s6_addr32[i]
		     && ip->s6_addr32[i] < map->first_ip.s6_addr32[i])
		 || (ip->s6_addr32[i] != map->last_ip.s6_addr32[i]
		     && ip->s6_addr32[i] > map->last_ip.s6_addr32[i]))
		     return 1;
	return 0;
}

static inline void
ip_netmask(const struct nf_set_ipmap *map, struct in6_addr *ip, struct in6_addr *masked)
{
	int i;
	
	for (i = 0; i < 4; i++)
		masked->s6_addr32[i] = 
			ip->s6_addr32[i] & map->netmask.s6_addr32[i];
}

static inline void
last_ip(struct in6_addr *to, struct in6_addr *from, struct in6_addr *mask)
{
	int i;
	
	for (i = 0; i < 4; i++)
		to->s6_addr32[i] = from->s6_addr32[i] | ~(mask->s6_addr32[i]);
}

static inline void
ip_from_skb(const struct sk_buff *skb, struct in6_addr *ip, int src)
{
	int i;
	
	ipv6_addr_copy(ip, src ? &skb->nh.ipv6h->saddr : &skb->nh.ipv6h->daddr);
	for (i = 0; i < 4; i++)
		ip->s6_addr32[i] = ntohl(ip->s6_addr32[i]);
}

static unsigned int
mask_to_bits(struct in6_addr *mask)
{
	unsigned int bits = 128;
	u_int32_t maskaddr;
	int i;
	
	for (i = 0; i < 4; i++) {
		if (mask->s6_addr32[i] == 0xFFFFFFFF)
			return bits;
	
		maskaddr = 0xFFFFFFFE;
		while (--bits >= (128 - 32*(i+1))) {
			maskaddr <<= 1;
			if (maskaddr != mask->s6_addr32[i])
				return bits;
		}
	}
	return bits;
}

#define ip_addr_copy(a,b)	ipv6_addr_copy(a,b)

static inline void
ip_dump(unsigned char *data, struct nf_set_ipmap *map, unsigned int i)
{
	struct in6_addr *ip = (struct in6_addr *)data;
	int bits = mask_to_bits(&map->netmask);
	
	ipv6_addr_copy(ip, &map->first_ip);
	ip->s6_addr32[bits % 32] += i*((128 - map->subnet) << 2);
	for (; i < 4; i++)
		ip->s6_addr32[i] = htonl(ip->s6_addr32[i]);
}

#else /* IPv4 */

static inline u_int32_t
ip_to_id(const struct nf_set_ipmap *map, u_int32_t *ip)
{
	return (*ip - map->first_ip) >> (32 - map->subnet);
}

static inline int
ip_out_of_range(const struct nf_set_ipmap *map, u_int32_t *ip)
{
	return (*ip < map->first_ip || *ip > map->last_ip);
}

static inline void
ip_netmask(const struct nf_set_ipmap *map, u_int32_t *ip, u_int32_t *masked)
{
	*masked = *ip & map->netmask;
}

static inline void
last_ip(u_int32_t *to, u_int32_t *from, u_int32_t *mask)
{
	*to = (*from) | ~(*mask);
}

static inline void
ip_from_skb(const struct sk_buff *skb, u_int32_t *ip, int src)
{
	*ip = ntohl(src ? skb->nh.iph->saddr : skb->nh.iph->daddr);
}

static unsigned int
mask_to_bits(u_int32_t *mask)
{
	unsigned int bits = 32;
	u_int32_t maskaddr;
	
	if (*mask == 0xFFFFFFFF)
		return bits;
	
	maskaddr = 0xFFFFFFFE;
	while (--bits >= 0 && maskaddr != *mask)
		maskaddr <<= 1;
	
	return bits;
}

#define ip_addr_copy(a,b)	*(b) = *(a)
#define ip_dump(data, map, i)	\
	*(u_int32_t *)data = map->first_ip + i * ((32 - map->subnet) << 2)

#endif /* IPv4 */

#endif /* __NF_SET_IPMAP_H */
