/* Copyright (C) 2000-2002 Joakim Axelsson <gozem@linux.nu>
 *                         Patrick Schaaf <bof@bof.de>
 * Copyright (C) 2003-2007 Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  
 */

/* The IPv4/IPv6 bitmap type */

#include <asm/bitops.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/string.h>

#include <linux/netfilter/nfnetlink.h>
#include <linux/netfilter/nf_set_netlink.h>
#include <linux/netfilter/nf_set.h>
#include <linux/netfilter/nf_set_ipmap.h>

static const int nfset_adt_ipmap_min[NFSET_ADT_IPMAP_MAX] = {
	[NFSET_ADT_IPMAP_IP-1] = sizeof(nf_set_ip_t),
	[NFSET_ADT_IPMAP_TIMEOUT-1] = 0,
};

static inline int
adt_data(struct nf_set_ipmap *map, struct nfattr *attr, nf_set_ip_t *ip)
{
	struct nfattr *tb[NFSET_ADT_IPMAP_MAX];
	nf_set_ip_t tmp, *maskedip = &tmp;
	
	nfattr_parse_nested(tb, NFSET_ADT_IPMAP_MAX, attr);
	if (nfattr_invalid_size(tb, NFSET_ADT_IPMAP_MAX, nfset_adt_ipmap_min))
		return -EINVAL;
	
	ip = (nf_set_ip_t *)NFA_DATA(tb[NFSET_ADT_IPMAP_IP-1]);

	if (ip_out_of_range(map, ip))
		return -ERANGE;

	ip_netmask(map, ip, maskedip);
	DSET_ADT(maskedip, ip);

	return 0;
}

static inline int
__testip(struct nf_set_ipmap *map, nf_set_ip_t *ip)
{	
	return !!test_bit(ip_to_id(map, ip), map->members);
}

static int
testip(struct nf_set *set, struct nfattr *attr)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;
	nf_set_ip_t *ip = NULL;
	int res = adt_data(map, attr, ip);

	if (res < 0)
		return res;

	return __testip(map, ip);
}

static int
testip_kernel(struct nf_set *set, 
	      const struct sk_buff *skb,
	      const u_int32_t flags)
{
	nf_set_ip_t tmp, *ip = &tmp;
	int res;
	
	ip_from_skb(skb, ip, flags & NFSET_DATA_FIRST);
	res =  __testip((struct nf_set_ipmap *) set->data, ip);
	return (res < 0 ? 0 : res);
}

static inline int
__addip(struct nf_set_ipmap *map, nf_set_ip_t *ip)
{
	if (test_and_set_bit(ip_to_id(map, ip), map->members))
		return -EEXIST;

	return 0;
}

static int
addip(struct nf_set *set, struct nfattr *attr)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;
	nf_set_ip_t *ip = NULL;
	int res = adt_data(map, attr, ip);

	if (res < 0)
		return res;

	return __addip(map, ip);
}

static int
addip_kernel(struct nf_set *set, 
	     const struct sk_buff *skb,
	     const u_int32_t flags)
{
	nf_set_ip_t tmp, *ip = &tmp;
	
	ip_from_skb(skb, ip, flags & NFSET_DATA_FIRST);
	return __addip((struct nf_set_ipmap *) set->data, ip);
}

static inline int 
__delip(struct nf_set_ipmap *map, nf_set_ip_t *ip)
{
	if (!test_and_clear_bit(ip_to_id(map, ip), map->members))
		return -EEXIST;
	
	return 0;
}

static int
delip(struct nf_set *set, struct nfattr *attr)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;
	nf_set_ip_t *ip = NULL;
	int res = adt_data(map, attr, ip);

	if (res < 0)
		return res;

	return __delip(map, ip);
}

static int
delip_kernel(struct nf_set *set,
	     const struct sk_buff *skb,
	     const u_int32_t flags)
{
	nf_set_ip_t tmp, *ip = &tmp;
	
	ip_from_skb(skb, ip, flags & NFSET_DATA_FIRST);
	return __delip((struct nf_set_ipmap *) set->data, ip);
}

static const int nfset_create_ipmap_min[NFSET_CREATE_IPMAP_MAX] = {
	[NFSET_CREATE_IPMAP_IP-1] = sizeof(nf_set_ip_t),
	[NFSET_CREATE_IPMAP_NETMASK-1] = sizeof(nf_set_ip_t),
	[NFSET_CREATE_IPMAP_SUBNET-1] = sizeof(u_int8_t),
	[NFSET_ADT_IPMAP_TIMEOUT-1] = 0,
};

static int create(struct nf_set *set, struct nfattr *attr)
{
	struct nf_set_ipmap *map;
	struct nfattr *tb[NFSET_CREATE_MAX];
	unsigned int size, netmask;

	nfattr_parse_nested(tb, NFSET_CREATE_MAX, attr);
	if (nfattr_invalid_size(tb, NFSET_CREATE_IPMAP_MAX, nfset_create_ipmap_min))
		return -EINVAL;
	map = kmalloc(sizeof(struct nf_set_ipmap), GFP_KERNEL);
	if (!map) {
		DSET("out of memory for %d bytes",
		   sizeof(struct nf_set_ipmap));
		return -ENOMEM;
	}
	
	ip_addr_copy(&map->first_ip,
		(nf_set_ip_t *) NFA_DATA(tb[NFSET_CREATE_IPMAP_IP-1]));
	ip_addr_copy(&map->netmask,
		(nf_set_ip_t *) NFA_DATA(tb[NFSET_CREATE_IPMAP_NETMASK-1]));

	/* Make sure it's the first IP */
	ip_netmask(map, &map->first_ip, &map->first_ip);
	last_ip(&map->last_ip, &map->first_ip, &map->netmask);
	
	map->subnet = *(u_int8_t *) NFA_DATA(tb[NFSET_CREATE_IPMAP_SUBNET-1]);
	
	/* Enforce size limit */
	netmask = mask_to_bits(&map->netmask);
	if (map->subnet <= netmask
	    || (map->subnet - netmask) > 16) {
	    	nf_set_printk("set %s: create set: invalid netmask (%d) and subnet mask (%d) combination",
	    		set->name, netmask, map->subnet);
		kfree(map);
		return -EINVAL;
	}
	
#define MAP_SIZE(map) (2 << (map->subnet - mask_to_bits(&map->netmask) - 1))

	DSET("set %s, netmask %d, size %u", 
		set->name, netmask, MAP_SIZE(map));
	size = bitmap_bytes(0, MAP_SIZE(map));
	map->members = kmalloc(size, GFP_KERNEL);
	if (!map->members) {
		DSET("out of memory for %d bytes", size);
		kfree(map);
		return -ENOMEM;
	}
	memset(map->members, 0, size);
	
	set->data = map;
	return 0;
}

static int destroy(struct nf_set *set)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;
	
	kfree(map->members);
	kfree(map);
	
	set->data = NULL;
	
	return 0;
}

static void flush(struct nf_set *set)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;
	memset(map->members, 0, bitmap_bytes(0, MAP_SIZE(map)));
}

static int list_header(struct sk_buff *skb, const struct nf_set *set)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;

	NFA_PUT(skb, NFSET_CREATE_IPMAP_IP, sizeof(map->first_ip),
		&map->first_ip);
	NFA_PUT(skb, NFSET_CREATE_IPMAP_NETMASK, sizeof(map->netmask),
		&map->first_ip);
	NFA_PUT(skb, NFSET_CREATE_IPMAP_SUBNET, sizeof(map->subnet),
		&map->subnet);
	
	return 0;
	
nfattr_failure:
	DSET("Should not happen. Ever. But we are here, confused, depressed...");
	
	return -EFAULT;
}

static int 
list_members(struct sk_buff *skb, 
	     struct netlink_callback *cb,
	     const struct nf_set *set)
{
	struct nf_set_ipmap *map = (struct nf_set_ipmap *) set->data;
	unsigned int i, size = MAP_SIZE(map);
	struct nfattr *nest_parms = NFA_NEST(skb, NFSET_ADT_DATA);
	unsigned char *data;

	/*  List as much as we can */
	for (i = cb->args[2]; i < size; i++)
		if (test_bit(i, map->members)) {
			if (skb_tailroom(skb) < NFA_ALIGN(sizeof(nf_set_ip_t)))
				break;
			data = skb_put(skb, NFA_ALIGN(sizeof(nf_set_ip_t)));
			ip_dump(data, map, i);
		}

	NFA_NEST_END(skb, nest_parms);
	
	if (i < size)
		cb->args[2] = i;
	else
		cb->args[1] = 0;

	return 0;

    nfattr_failure:
    	return -EINVAL;
}

struct nf_set_type nf_set_type_ipmap = {
	.typename		= "ipmap",
	.features		= NFSET_TYPE_IP | NFSET_DATA_FIRST,
	.protocol_version	= NF_SET_PROTOCOL_VERSION,
	.family			= AF_FAMILY,
	.create			= &create,
	.destroy		= &destroy,
	.flush			= &flush,
	.addip			= &addip,
	.addip_kernel		= &addip_kernel,
	.delip			= &delip,
	.delip_kernel		= &delip_kernel,
	.testip			= &testip,
	.testip_kernel		= &testip_kernel,
	.list_header		= &list_header,
	.list_members		= &list_members,
	.me			= THIS_MODULE,
};
